/*
 * adl_control.c -- a silly tool to send ioctl to the raw device
 *
 * Copyright (C) 1998   rubini@linux.it (Alessandro Rubini)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "adl7200.h"

struct command {
    char *name;             /* symbolic name for cmdline interaction */
    char *macro;            /* actual name */
    unsigned int cmd;       /* ioctl number */
    unsigned int nargs;     /* number of arguments (numeric) */
    int (*fun)(int, int, unsigned long *, char *);
};

/* ---------------- functions in charge of dealing with the command */

int do_long(int fd, int cmd, unsigned long *args, char *outptr)
{
    int retval = ioctl(fd,cmd,args);

    if (retval < 0) {
	return retval;
    }

    if (_IOC_DIR(cmd) & _IOC_READ)
	    sprintf(outptr, "\"0x%08lx (%li)\" ",args[0],args[0]);
    return retval;
}
    

/* ------------------------------------------------------ the table */
struct command commands[] = {
    {"reset",      "ADL_IOCRESET",     ADL_IOCRESET,     0, NULL},
    {"hardreset",  "ADL_IOCHARDRESET", ADL_IOCHARDRESET, 0, NULL},
    {"getflags",   "ADL_IOCGFLAGS",    ADL_IOCGFLAGS,    0, do_long},
    {"setflags",   "ADL_IOCSFLAGS",    ADL_IOCSFLAGS,    1, do_long},
    {"gettimer",   "ADL_IOCGTIMER",    ADL_IOCGTIMER,    0, do_long},
    {"settimer",   "ADL_IOCSTIMER",    ADL_IOCSTIMER,    1, do_long},
    {"gethertz",   "ADL_IOCGHERTZ",    ADL_IOCGHERTZ,    0, do_long},
    {"sethertz",   "ADL_IOCSHERTZ",    ADL_IOCSHERTZ,    1, do_long},
    {"getdmasize", "ADL_IOCGDMASIZE",  ADL_IOCGDMASIZE,  0, do_long},
    {"setdmasize", "ADL_IOCSDMASIZE",  ADL_IOCSDMASIZE,  1, do_long},
    {"getdmacount","ADL_IOCGDMACOUNT", ADL_IOCGDMACOUNT, 0, do_long},
    {"getunderrun","ADL_IOCGUNDERRUN", ADL_IOCGUNDERRUN, 0, do_long},
    {"getoverrun", "ADL_IOCGOVERRUN",  ADL_IOCGOVERRUN,  0, do_long},
    
    {NULL,0,0}
    };



int main(int argc, char **argv)
{
    FILE *f=NULL;
    char *device;
    struct command *ptr;
    int i, j, cmd, retval;
    unsigned long nums[4];
    char trash[24];
    char output[80];

    if (argc<2) {
        fprintf(stderr,"Use: %s [command [arg] ...]\n",argv[0]);
	fprintf(stderr,"The device used is stdout, or $ADLDEVICE\n");
	fprintf(stderr,"Commands are:\n");
	for (ptr=commands; ptr->name; ptr++) {
	    fprintf(stderr,"  %-15s (%-20s, %2i numeric arg%s)\n",
		    ptr->name, ptr->macro, ptr->nargs,
		    ptr->nargs == 1 ? "" : "s");
	}
        exit(1);
    }
    device=getenv("ADLDEVICE");
    if (!device) {device="<stdout>"; f=stdout;}
    if (!f)
        f=fopen(device,"r+");
    if (!f) {
        fprintf(stderr,"%s: %s: %s\n",argv[0],device,strerror(errno));
        exit(1);
    }
    for (i=1; i<argc; i++) {
	for (ptr=commands; ptr->name; ptr++) {
	    if (strcmp(ptr->name,argv[i]) && strcmp(ptr->macro,argv[i]))
		continue;

	    /* found */
	    cmd = ptr->cmd;
	    for (j=0; j<ptr->nargs; j++) {
		if (!argv[i+1+j]) {
		    fprintf(stderr,"%s: too few arguments for \"%s\"\n",
			    argv[0],ptr->name);
		    exit(1);
		}
		if (sscanf(argv[i+1+j],"%li%s",&nums[j],trash)!=1) {
		    fprintf(stderr,"%s: not a number \"%s\"\n",
			    argv[0],argv[i+1+j]);
		    exit(1);
		}
	    }
	    i+=j;
	    output[0] = '\0'; /* no output */
	    if (ptr->fun) /* do that via function */
		retval = ptr->fun(fileno(f),cmd,nums,output);
	    else
		retval = ioctl(fileno(f),cmd,nums);
	    fprintf(stderr,"ioctl(\"%s\", %s, ...) = %s%i", device,
		   ptr->macro, output, retval);
	    if (retval < 0)
		fprintf(stderr," (%s)",strerror(errno));
	    fprintf(stderr,"\n");
	    break;
	}
	if (!ptr->name) {
	    fprintf(stderr,"%s: invalid command \"%s\"\n",argv[0],argv[1]);
	    exit(1);
	}
    }

    fclose(f);
    exit(0);
}
