/*
 * dma.c -- dma support: imlementation of /dev/adlD7200 and /dev/adlT7200
 *
 * Copyright (C) 1998   rubini@linux.it (Alessandro Rubini)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__
#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/bios32.h> /* pcibios_* */
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>

#include "sysdep-2.1.h"
#include "adl7200.h"        /* local definitions */

/*--------------------------------------------------------------- lseek */

/* That's easy, as I'm lazy :) */

lseek_t adl_dma_lseek(struct inode *inode, struct file *file,
		      lseek_off_t offset, int whence)
{
    return -ESPIPE;
}

/*--------------------------------------------------------------- open */
int adl_dma_open (struct inode *inode, struct file *filp)
{
    return 0;
}


/*--------------------------------------------------------------- close */

release_t adl_dma_close (struct inode *inode, struct file *filp)
{
#if 0
    Adl_Dev *dev = filp->private_data;

    if (dev->flags & ADL_FLAG_WRITING) { /* busy */
        interruptible_sleep_on(&dev->closeq);
    }
#endif

    PDEBUG("dma close\n");
    /* fall back on the other one */
    adl_fops[0]->release(inode,filp);
    release_return(0);
}

/*-------------------------------------------------------- read */

read_write_t adl_dma_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;
    struct wait_queue wait = { current, NULL };
    count_t newcount;

    if (dev->flags & ADL_FLAG_WRITING)
	return -EBUSY;

    if (filp->f_pos == 0 && !(dev->flags & ADL_FLAG_READING)) {
	/* first read: allocate and trigger */
        adl_allocate_dma(dev,(dev->dma_requested_size+1023) >> 10,
			 GFP_KERNEL);
        if (!dev->dmabuffer)
            return -ENOMEM;

	cli();
	dev->rpending = dev->dma_requested_size;
	dev->flags |= ADL_FLAG_READING;
	sti();

	/*
	 * Now, setup things according to device type: if this is
	 * /dev/adlD, then activate the timer and disable I_REQ,
	 * if this is /dev/adlT do the other way round.
	 */
	if (TYPE(inode->i_rdev) == ADL_DMA) {
	    adl_run_dma(dev, ADL_BIT_IENA |
			ADL_BIT_IFIF | ADL_BIT_ICNT,   /* bits to set */
			ADL_BIT_IREQ | ADL_BIT_IACK);  /* bits to clear */
	    adl_set_clock(dev);
	} else {
	    adl_run_dma(dev, ADL_BIT_IENA | ADL_BIT_IACK |
			ADL_BIT_IFIF | ADL_BIT_IREQ,   /* bits to set */
			ADL_BIT_ICNT);                 /* bits to clear */
	    /* no need to shut the clock, as ICNT is disabled */
	}

	if (dev->hertz) { /* prepare for timely awake */
	    del_timer(&dev->timer);
	    dev->timer.expires = jiffies+HZ/dev->hertz;
	    add_timer(&dev->timer);
	}

    }

    /* at this point, either DMA is running, or data is there */

    newcount = dev->dma_requested_size - dev->rpending - filp->f_pos;
    if (!newcount) { /* return or sleep */
	if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;

	/* now go to sleep in a safe way */
	add_wait_queue(&dev->inq, &wait);
	do {
            current->state = TASK_INTERRUPTIBLE;
            schedule();
        }
        while (dev->rpending && \
               (filp->f_pos == dev->dma_requested_size - dev->rpending) &&
               !(current->signal & ~current->blocked));
        remove_wait_queue(&dev->inq, &wait);
        if (current->signal & ~current->blocked) /* a signal arrived */
            return -ERESTARTSYS;
        PDEBUGG("read awake: pending is %li\n",dev->rpending);
    }

    newcount = dev->dma_requested_size - dev->rpending - filp->f_pos;
    if (!newcount) {
	printk(KERN_WARNING ADL_MSG "read: awake with no data\n");
	return -ENODEV; 
    }

    if (count > newcount) {
	count = newcount;
    }

    __copy_to_user(buf,(void *)dev->dmaremap+filp->f_pos, count);
    filp->f_pos += count;
    if (filp->f_pos == dev->dma_requested_size)
	filp->f_pos = 0;
    return count;
}

/*-------------------------------------------------------- write */

read_write_t adl_dma_write (struct inode *inode, struct file *filp,
                const char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;
    struct wait_queue wait = { current, NULL };

    if (dev->flags & ADL_FLAG_READING)
	return -EBUSY;

    if ((dev->flags & ADL_FLAG_WRITING)
	&& dev->wpending &&
	(filp->f_pos == dev->dma_requested_size - dev->wpending)) {


	PDEBUGG("write: go to sleep (pos %08lx, pending %08lx, size %08x\n",
	       (long)(filp->f_pos), dev->wpending, dev->dma_requested_size);
	/* go to sleep, and do it in the safe way */
        if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;
	add_wait_queue(&dev->outq, &wait);
	do {
            current->state = TASK_INTERRUPTIBLE;
	    schedule();
        }
	while (dev->wpending && \
	       (filp->f_pos == dev->dma_requested_size - dev->wpending) &&
               !(current->signal & ~current->blocked));
	remove_wait_queue(&dev->outq, &wait);
        if (current->signal & ~current->blocked) /* a signal arrived */
            return -ERESTARTSYS;
        PDEBUGG("awake: pending is %li\n",dev->wpending);
    }
    PDEBUGG("write: pending %li\n",dev->wpending);
    if (filp->f_pos >= dev->dma_requested_size) {
        filp->f_pos = 0;
    }

    if (filp->f_pos == 0) { /* first write: allocate */
        adl_allocate_dma(dev,(dev->dma_requested_size+1023) >> 10,
			 GFP_KERNEL);
        if (!dev->dmabuffer) 
            return -ENOMEM;
    }

    if (filp->f_pos + count > dev->dma_requested_size - dev->wpending) {
        count = dev->dma_requested_size - dev->wpending - filp->f_pos;
    }

    __copy_from_user((void *)(dev->dmaremap+(unsigned long)filp->f_pos),
			      buf, count);
    filp->f_pos += count;

    if (filp->f_pos < dev->dma_requested_size)
        return count; /* not enough data, yet */

    /* If allowed, activate the monitoring timer */
    if (dev->hertz) {
        del_timer(&dev->timer);
	dev->timer.expires = jiffies+HZ/dev->hertz;
	add_timer(&dev->timer);
    }

    /* Ok, now setup DMA */
    cli();
    dev->wpending = dev->dma_requested_size;
    dev->flags |= ADL_FLAG_WRITING;
    sti();
    adl_run_dma(dev, 
                ADL_BIT_OFIF | ADL_BIT_OCNT,   /* bits to set */
                ADL_BIT_OREQ | ADL_BIT_OACK);  /* bits to clear */
    adl_set_clock(dev);

    return count;
}

/*-------------------------------------------------------- fops */

struct file_operations adl_dma_fops = {
    adl_dma_lseek,
    adl_dma_read,
    adl_dma_write,
    NULL,            /* adl_readdir */
    adl_poll,
    adl_ioctl,
    NULL,            /* adl_mmap */
    adl_dma_open,
    adl_dma_close,
    NULL,          /* nothing more... */
};


struct file_operations adl_trg_fops = {
    NULL,          /* adl_lseek */
    adl_dma_read,
    adl_dma_write,
    NULL,          /* adl_readdir */
    adl_poll,
    adl_ioctl,
    NULL,          /* adl_mmap */
    adl_dma_open,
    adl_dma_close,
    NULL,          /* adl_fsync */
    NULL,          /* adl_fasync */
                   /* nothing more, fill with NULLs */
};
